local atan2,sqrt,clamp,random = math.atan2,math.sqrt,math.clamp,math.random
local INF = math.huge

-- local functions
local function smoothOver(dt, smoothTime, convergenceFraction) return 1 - (1 - convergenceFraction)^(dt / smoothTime) end
local function lerp(a,b,t) return a + (b-a) * t end
local function distanceSq(x1,y1, x2,y2) return sqrt((x2-x1)^2 + (y2-y1)^2) end
local function angle(x1,y1, x2,y2) return atan2(y2-y1,x2-x1) end
local function outExponential(ratio) if ratio == 1 then return 1 end return 1-2^(-10 * ratio) end

-- GCam class
GCam = Core.class(Sprite)
GCam.SHAKE_DELAY = 10

function GCam:init(content, ax, ay)
	assert(content ~= stage, "bad argument #1 (content should be different from the 'stage')")
	self.viewport = Viewport.new()
	self.viewport:setContent(content)

	self.content = Sprite.new()
	self.content:addChild(self.viewport)
	self:addChild(self.content)

	self.matrix = Matrix.new()
	self.viewport:setMatrix(self.matrix)
	
	self.w = 0
	self.h = 0
	self.ax = ax or 0.5
	self.ay = ay or 0.5
	self.x = 0
	self.y = 0
	self.zoomFactor = 1
	self.rotation = 0

	-- Bounds
	self.leftBound   = -INF
	self.rightBound  = INF
	self.topBound    = -INF
	self.bottomBound = INF
	
	-- Shaker
	self.shakeTimer = Timer.new(GCam.SHAKE_DELAY, 1)
	self.shakeDistance = 0
	self.shakeCount = 0
	self.shakeAmount = 0
	self.shakeTimer:addEventListener("timerComplete", self.shakeDone, self)
	self.shakeTimer:addEventListener("timer", self.shakeUpdate, self)
	self.shakeTimer:stop()
	
	-- to detect object velocity
	self.__prevPosX = 0
	self.__prevPosY = 0
	
	-- Follow
	-- 0 - instant move
	self.smoothX = 0.9
	self.smoothY = 0.9
	-- Dead zone
	self.deadWidth  = 50
	self.deadHeight = 50
	self.deadHeightOffset = 0 -- XXX new 20220918
	-- Soft zone
	self.softWidth  = 150
	self.softHeight = 150
	
	---------------------------------------
	------------- debug stuff -------------
	---------------------------------------
	self.__debugSoftColor = 0xffff00
	self.__debugAnchorColor = 0xff0000
	self.__debugDotColor = 0x00ff00
	self.__debugAlpha = 0.5
	
	self.__debugRMesh = Mesh.new()
	self.__debugRMesh:setIndexArray(
		1,3,4, 1,2,4, 1,3,7, 3,5,7, 2,4,8, 4,8,6, 5,6,8, 5,8,7, 9,10,11, 9,11,12, 13,14,15, 13,15,16, 17,18,19, 17,19,20
	)
	self.__debugRMesh:setColorArray(
		self.__debugSoftColor,self.__debugAlpha, self.__debugSoftColor,self.__debugAlpha, self.__debugSoftColor,self.__debugAlpha,
		self.__debugSoftColor,self.__debugAlpha,  self.__debugSoftColor,self.__debugAlpha, self.__debugSoftColor,self.__debugAlpha,
		self.__debugSoftColor,self.__debugAlpha, self.__debugSoftColor,self.__debugAlpha, self.__debugAnchorColor,self.__debugAlpha,
		self.__debugAnchorColor,self.__debugAlpha, self.__debugAnchorColor,self.__debugAlpha, self.__debugAnchorColor,self.__debugAlpha,
		self.__debugAnchorColor,self.__debugAlpha, self.__debugAnchorColor,self.__debugAlpha, self.__debugAnchorColor,self.__debugAlpha,
		self.__debugAnchorColor,self.__debugAlpha, self.__debugDotColor,self.__debugAlpha, self.__debugDotColor,self.__debugAlpha,
		self.__debugDotColor,self.__debugAlpha, self.__debugDotColor,self.__debugAlpha
	)
	---------------------------------------
	---------------------------------------
	---------------------------------------

	self:setAnchor(self.ax,self.ay)
end

---------------------------------------------------
--------------------- FOLLOW ----------------------
---------------------------------------------------
function GCam:setFollow(obj)
	self.__prevPosX, self.__prevPosY = obj:getPosition()	

	self.followObj = obj
end

---------------------------------------------------
---------------------- SHAPES ---------------------
---------------------------------------------------
function GCam:rectangle(dt, x, y)
	local dw = self.deadWidth  / 2
	local dh = self.deadHeight / 2
	local sw = self.softWidth  / 2
	local sh = self.softHeight / 2
	
	local dstX = self.x
	local dstY = self.y
	
	-- X smoothing
	if x > self.x + dw then -- out of dead zone on right side
		local t = smoothOver(dt, self.smoothX, 0.99)
		local newX = lerp(self.x, x - dw, t)
		dstX = clamp(newX, x - sw, x + sw)
	elseif x < self.x - dw then  -- out of dead zone on left side
		local t = smoothOver(dt, self.smoothX, 0.99)
		local newX = lerp(self.x, x + dw, t)
		dstX = clamp(newX, x - sw, x + sw)
	end
	-- clamp to soft zone
	
	-- Y smoothing
	if y < self.y - dh - self.deadHeightOffset then  -- out of dead zone on top side XXX new 20220918
		local t = smoothOver(dt, self.smoothY, 0.99) -- 0.99
		local newY = lerp(self.y, y + dh + self.deadHeightOffset, t)
--		dstY = clamp(newY, y-sh, y+sh)
		dstY = clamp(newY, y-sh + self.deadHeightOffset, y+sh + self.deadHeightOffset)
	elseif y > self.y + dh + self.deadHeightOffset then -- out of dead zone on bottom side
		local t = smoothOver(dt, self.smoothY, 0.99) -- 0.99
		local newY = lerp(self.y, y - dh - self.deadHeightOffset, t)
--		dstY = clamp(newY, y-sh, y+sh)
		dstY = clamp(newY, y-sh - self.deadHeightOffset, y+sh - self.deadHeightOffset)
	end
	-- clamp to soft zone
	
	return dstX, dstY
end

---------------------------------------------------
---------------------- UPDATE ---------------------
---------------------------------------------------
function GCam:update(dt)
	local x,y = self.followObj:getPosition()

	self.__prevPosX = x
	self.__prevPosY = y

	local dstX, dstY = self:rectangle(dt, x, y)

	if self.x ~= dstX or self.y ~= dstY then self:goto(dstX, dstY) end

	-- debug
	if self.__debug__ then self:debugUpdate(true, x, y) end
end

------------------------------------------
---------------- POSITION ----------------
------------------------------------------
function GCam:rawGoto(x,y)
	x, y = self:checkBounds(x,y)
	
	self.matrix:setAnchorPosition(x,y)
	self.viewport:setMatrix(self.matrix)	
end

function GCam:goto(x,y)
	self.x, self.y = self:checkBounds(x,y)
	
	self.matrix:setAnchorPosition(self.x, self.y)
	self.viewport:setMatrix(self.matrix)
end

function GCam:gotoX(x)
	self.x = self:checkBounds(x)
	
	self.viewport:set("anchorX", self.x)
end

function GCam:gotoY(y)
	self.y = self:checkBoundY(y)

	self.viewport:set("anchorY", self.y)
end

--------------------------------------------------
--------------------- BOUNDS ---------------------
--------------------------------------------------
function GCam:checkBounds(x,y)
	local offX = self.w * self.ax
	local offY = self.h * self.ay
	local newX = clamp(x, self.leftBound + offX, self.rightBound - offX)
	local newY = clamp(y, self.topBound + offY, self.bottomBound - offY)
	return newX,newY
end

function GCam:checkBoundX(x)
	local offX = self.w * self.ax
	local newX = clamp(x, self.leftBound + offX, self.rightBound - offX)
	return newX
end

function GCam:checkBoundY(y) -- new 20230515 XXX
	local offY = self.h * self.ay
	local newY = clamp(y, self.topBound + offY, self.bottomBound - offY)
	return newY
end

function GCam:updateBounds()
	local x,y = self:checkBounds(self.x, self.y)
	if x ~= self.x or y ~= self.y then 
		self:goto(x,y)
	end
end

-- Camera can move only inside given box
function GCam:setBounds(left, top, right, bottom)
	self.leftBound = left or 0
	self.topBound = top or 0
	self.rightBound = right or 0
	self.bottomBound = bottom or 0
	
	self:updateBounds()
end

function GCam:setLeftBound(left)
	self.leftBound = left or 0
	self:updateBounds()
end

function GCam:setTopBound(top)
	self.topBound = top or 0
	self:updateBounds()
end

function GCam:setRightBound(right)
	self.rightBound = right or 0
	self:updateBounds()
end

function GCam:setBottomBound(bottom)
	self.bottomBound = bottom or 0
	self:updateBounds()
end

function GCam:getBounds() 
	return self.leftBound, self.topBound, self.rightBound, self.bottomBound
end

--------------------------------------------------
--------------------- ZONES ----------------------
--------------------------------------------------
--	Camera does not move in dead zone
-- w (number): dead zone width
-- h (number): dead zone height
--function GCam:setDeadSize(w, h)
function GCam:setDeadSize(w, h, o)
	self.deadWidth = w
	self.deadHeight = h or w
	self.deadHeightOffset = o or 0
	self:debugUpdate()
end

function GCam:setDeadWidth(w)
	self.deadWidth = w
	self:debugUpdate()
end

function GCam:setDeadHeight(h)
	self.deadHeight = h
	self:debugUpdate()
end

--	Camera intepolate its position towards target
-- w (number): soft zone width
-- h (number): soft zone height
function GCam:setSoftSize(w, h)
	self.softWidth = w
	self.softHeight = h or w
	self:debugUpdate()
end

function GCam:setSoftWidth(w)
	self.softWidth = w
	self:debugUpdate()
end

function GCam:setSoftHeight(h)
	self.softHeight = h
	self:debugUpdate()
end

-- Smooth factor
--	x (number):
--	y (number):
function GCam:setSmooth(x, y)
	self.smoothX = x
	self.smoothY = y or x
end

function GCam:setSmoothX(x)
	self.smoothX = x
end

function GCam:setSmoothY(y)
	self.smoothY = y
end

---------------------------------------------------
----------------- TRANSFORMATIONS -----------------
---------------------------------------------------
function GCam:move(dx, dy)
	self:goto(self.x + dx, self.y + dy)
end

function GCam:zoom(value)
	local v = self.zoomFactor + value
	if v > 0 then self:setZoom(v) end
end

function GCam:rotate(ang)
	self.rotation += ang
	self:setAngle(self.rotation)
end

------------------------------------------
------------------ ZOOM ------------------
------------------------------------------
function GCam:setZoom(zoom)
	self.zoomFactor = zoom
	self.matrix:setScale(zoom, zoom, 1)
	self.viewport:setMatrix(self.matrix)
	self:debugUpdate()
end

function GCam:getZoom()
	return self.zoomFactor
end

------------------------------------------
---------------- ROTATION ----------------
------------------------------------------
function GCam:setAngle(angle)
	self.rotation = angle
	self.matrix:setRotationZ(angle)
	self.viewport:setMatrix(self.matrix)
	self:debugUpdate()
end

function GCam:getAngle()
	return self.matrix:getRotationZ()
end

---------------------------------------------------
---------------------- SHAKE ----------------------
---------------------------------------------------
-- duration (number): time is s.
-- distance (number): maximum shake offset
function GCam:shake(duration, distance)
	self.shaking = true
	
	self.shakeCount = 0
	self.shakeDistance = distance or 100
	self.shakeAmount = (duration*1000) // GCam.SHAKE_DELAY
	
	self.shakeTimer:reset()
	self.shakeTimer:setRepeatCount(self.shakeAmount)
	self.shakeTimer:start()
end

function GCam:shakeDone()
	self.shaking = false
	self.shakeCount = 0
	self.content:setPosition(0,0)
end

function GCam:shakeUpdate()
	self.shakeCount += 1
	local amplitude = 1 - outExponential(self.shakeCount/self.shakeAmount)
	local hd = self.shakeDistance / 2
	local x = random(-hd,hd)*amplitude
	local y = random(-hd,hd)*amplitude
	self.content:setPosition(x, y)
end

------------------------------------------
-------------- ANCHOR POINT --------------
------------------------------------------
function GCam:setAnchor(anchorX, anchorY)
	self.ax = anchorX
	self.ay = anchorY
	self.matrix:setPosition(self.w * anchorX,self.h * anchorY)
	self.viewport:setMatrix(self.matrix)
	self:debugUpdate()
end

function GCam:setAnchorX(anchorX)
	self.ax = anchorX
	self.matrix:setPosition(self.w * anchorX,self.h * self.ay)
	self.viewport:setMatrix(self.matrix)
	self:debugUpdate()
end

function GCam:setAnchorY(anchorY)
	self.ay = anchorY
	self.matrix:setPosition(self.w * self.ax,self.h * anchorY)
	self.viewport:setMatrix(self.matrix)
	self:debugUpdate()
end

function GCam:getAnchor()
	return self.ax, self.ay
end
function GCam:getAnchorX()
	return self.ax
end
function GCam:getAnchorY()
	return self.ay
end

---------------------------------------------------
------------------- DEBUG STUFF -------------------
---------------------------------------------------
function GCam:setDebug(flag)
	self.__debug__ = flag

	if flag then 
		self:addChild(self.__debugRMesh)
		self:debugUpdate()
		self:debugUpdate(true, 0, 0)
	else
		self.__debugRMesh:removeFromParent()
	end
end

function GCam:switchDebug()
	self:setDebug(not self.__debug__)
end

function GCam:debugMeshUpdate()
	local w,h = self.w, self.h
	local zoom = self.zoomFactor
	local rot = self.rotation
	
	local ax,ay = w*self.ax, h*self.ay
	
	local TS = 1
	local off = w <> h
	
	local dw = (self.deadWidth  * zoom) / 2
	local dh = (self.deadHeight * zoom) / 2
	local sw = (self.softWidth  * zoom) / 2
	local sh = (self.softHeight * zoom) / 2
	--[[
	Mesh vertices
	1-----------------2
	| \  soft zone  / |
	|  3-----------4  |
	|  | dead zone |  |
	|  5-----------6  |
	| /             \ |
	7-----------------8
	]]	
	self.__debugRMesh:setVertexArray(
		ax-sw,ay-sh, -- 1
		ax+sw,ay-sh, -- 2

--		ax-dw,ay-dh, -- 3
--		ax+dw,ay-dh, -- 4
		ax-dw,ay-dh+self.deadHeightOffset, -- 3 XXX new 20220918
		ax+dw,ay-dh+self.deadHeightOffset, -- 4 XXX new 20220918
--		ax-dw,ay+dh, -- 5
--		ax+dw,ay+dh, -- 6
		ax-dw,ay+dh+self.deadHeightOffset, -- 5 XXX new 20230515
		ax+dw,ay+dh+self.deadHeightOffset, -- 6 XXX new 20230515

--		ax-sw,ay+sh, -- 7
--		ax+sw,ay+sh, -- 8
		ax-sw,ay+sh+self.deadHeightOffset, -- 7 XXX new 20220918
		ax+sw,ay+sh+self.deadHeightOffset, -- 8 XXX new 20220918

		ax-TS,-off, ax+TS,-off, -- vertical line
		ax+TS,h+off, ax-TS,h+off, -- vertical line

		-off,ay-TS, -off,ay+TS, -- horizontal line
		w+off,ay+TS, w+off,ay-TS -- horizontal line
	)
	self.__debugRMesh:setAnchorPosition(ax,ay)
	self.__debugRMesh:setPosition(ax,ay)
	self.__debugRMesh:setRotation(rot)
end

function GCam:debugUpdate(dotOnly, gx, gy)
	if self.__debug__ then
		if dotOnly then -- a green little pixel :-)
			local zoom = self:getZoom()
			local ax = self.w * self.ax
			local ay = self.h * self.ay
			local size = 16 * zoom

			local x = (gx * zoom - self.x * zoom) + ax
			local y = (gy * zoom - self.y * zoom) + ay
			self.__debugRMesh:setVertex(17, x-size,y-size)
			self.__debugRMesh:setVertex(18, x+size,y-size)
			self.__debugRMesh:setVertex(19, x+size,y+size)
			self.__debugRMesh:setVertex(20, x-size,y+size)
		else
			self:debugMeshUpdate()
		end
	end
end

------------------------------------------
------------------ SIZE ------------------
------------------------------------------
function GCam:setSize(w,h)
	self.w = w
	self.h = h

	self:debugUpdate()
end

---------------------------------------------------
----------------- RESIZE LISTENER -----------------
---------------------------------------------------
-- set camera size to window size
function GCam:setAutoSize(flag)
	if flag then
		self:addEventListener(Event.APPLICATION_RESIZE, self.appResize, self)
		self:appResize()
	elseif self:hasEventListener(Event.APPLICATION_RESIZE) then
		self:removeEventListener(Event.APPLICATION_RESIZE, self.appResize, self)
	end
end

function GCam:appResize()
	local minX,minY,maxX,maxY = application:getLogicalBounds()
	self.w = maxX+minX
	self.h = maxY+minY
	self.matrix:setPosition(self.w*self.ax, self.h*self.ay)
	self.viewport:setMatrix(self.matrix)

	self:debugUpdate()
end
